from collections import defaultdict

from flask import current_app, request
from flask.helpers import _endpoint_from_view_func
from werkzeug.local import LocalProxy
from werkzeug.exceptions import NotAcceptable

current_blueprint = LocalProxy(
    lambda: current_app.blueprints[request.blueprint])


def register_converter(app_or_blueprint, name, converter):
    """Registers a converter on a url_map of an app or adds it to a blueprint.
    """

    # the following is a helper to register a converter through a blueprint
    def _register_converter(app):
        app.url_map.converters[name] = converter

    # attach directly if called on app, through blueprint otherwise
    if hasattr(app_or_blueprint, 'url_map'):
        _register_converter(app_or_blueprint)
    else:
        app_or_blueprint.record_once(
            lambda state: _register_converter(state.app))


def serialize_response(response_data,
                       content_type=None,
                       status=200,
                       renderer=None):
    """Serializes a response using a specified renderer.

    This will serialize ``response_data`` with the specified ``content_type``,
    using ``renderer``.

    If ``content_type`` is ``None``,
    :py:func:`~flask_arrest.helpers.get_best_mimetype` will be used to
    determine a suitable type. If no match is found, a
    :py:class:`~werkzeug.exceptions.NotAcceptable` exception is thrown.

    If no render is supplied, use the blueprint's
    :attr:`~flask_arrest.RestBlueprint.content_renderer`.

    :param response_data: Data to be serialized. Can be anything the renderer
                          can handle.
    :param content_type: The ``Content-type`` to serialize for.
    :param status: The response status to pass on to the renderer.
    :param renderer: The renderer to use. If ``None``, lookup the current
                     blueprint's
                     :attr:`~flask_arrest.RestBlueprint.content_renderer`.
    :return: A :class:`~flask.Response` object.

    """
    content_type = content_type or get_best_mimetype()

    if not content_type:
        # no accepted content-type. send flasks default 406, instead of raising
        return NotAcceptable()

    if not renderer:
        renderer = current_blueprint.content_renderer
    return renderer.render_response(response_data, content_type, status)


def get_best_mimetype():
    """Returns the highest quality server-to-client content-type that both
    agree on. Returns ``None``, if no suitable type is found.

    Internally, works by querying the blueprint for its
    :attr:`~flask_arrest.ContentNegotiationMixin.outgoing` attribute and
    comparing it with the ``Accept``-headers sent by the client.."""
    # find out what the client accepts
    return request.accept_mimetypes.best_match(
        current_blueprint.outgoing.get_mimetypes(request.endpoint))


class MIMEMap(object):
    """Special datastructure that maps an endpoint to a set of mimetypes. The
    default set of mimetypes for any endpoint is ``{None}``.

    A value of ``None`` will be replaced with all mimetypes set for the
    special endpoint
    :py:const:`~flask_arrest.helpers.MIMEMap.DEFAULT_ENDPOINT`.

    If ``None`` is *not* included in a specific endpoints set of mimetypes, it
    will not include the defaults mentioned above.

    Note that endpoint names should usually be added without the
    Blueprint-prefix (i.e. "index" instead of "api.index").
    """

    #: The default endpoint. Any value of ``None`` in the set of acceptable
    #: types in other endpoints will be replaced with all types assigned to
    #: this endpoint.
    DEFAULT_ENDPOINT = None

    def __init__(self):
        self._map = defaultdict(lambda: set([None]))

    def add_mimetype(self, mimetype, endpoint=DEFAULT_ENDPOINT):
        """Adds a mimetype to an endpoint."""
        if mimetype is None and endpoint is self.DEFAULT_ENDPOINT:
            raise ValueError('Cannot add default mimetype on default.')

        self._map[endpoint].add(mimetype)

    def set_mimetypes(self, mimetypes, endpoint=DEFAULT_ENDPOINT):
        """Sets all mimetypes for an endpoint.

        Note that if you exclude ``None`` from the ``mimetypes`` set, the
        mimetypes set for
        :py:const:`~flask_arrest.helpers.MIMEMap.DEFAULT_ENDPOINT` will not be
        applied. This allows specifying an endpoint that does not handle all
        of the otherwise common mimetypes."""
        if endpoint is self.DEFAULT_ENDPOINT and None in mimetypes:
            raise ValueError('Cannot include default mimetype in default.')

        self._map[endpoint] = set(mimetypes)

    def get_mimetypes(self, endpoint=DEFAULT_ENDPOINT):
        """Get all mimetypes for an endpoint."""
        mimetypes = set(self._map[endpoint])

        if None in mimetypes:
            mimetypes.update(self._map[None])
            mimetypes.remove(None)

        return mimetypes

    def add(self, extra_type):
        def _(f):
            self.add_mimetype(extra_type, _endpoint_from_view_func(f))
            return f

        return _

    def only(self, only_types):
        def _(f):
            self.set_mimetypes(only_types, _endpoint_from_view_func(f))
            return f

        return _



from __future__ import absolute_import

from copy import deepcopy
import json
from pprint import pformat

from flask import make_response, current_app
from .helpers import current_blueprint
import jsonext


# encoding to be used when sending text/plain
TEXT_PLAIN_ENCODING = 'utf8'


class Renderer(object):
    """Basic Renderer interface.

    A Renderer can be asked to render an an object into a response."""
    def render_response(self, data, content_type, status=200):
        """Render data.

        :param data: The data to be rendered.
        :param content_type: A string describing the desired output
                             content-type.
        :param status: The status code for the response.
        :return: A :class:`~flask.Response` instance."""
        raise NotImplementedError


class PluggableRenderer(Renderer):
    """Support rendering content by registering rendering functions for each
    content type.

    Any renderer will be called with arguments matching ``data, content_type``,
    where ``data`` is the object to be rendered and ``content_type`` the
    desired content-type as a string. The return value is passed as arguments
    to :func:`~flask.make_response`.
    """
    def __init__(self, *args, **kwargs):
        super(PluggableRenderer, self).__init__(*args, **kwargs)
        self.content_funcs = {}

    def register_renderer(self, content_type, func):
        """Set renderer for ``content_type`` to func."""
        self.content_funcs[content_type] = func

    def renders(self, content_type):
        """A function decorator. Decorating a function with this is equivalent
        to calling ``register_renderer(content_type, this_function)``.
        """
        def _(f):
            self.register_renderer(content_type, f)
            return f
        return _

    def render_response(self, data, content_type, status=200):
        if not content_type in self.content_funcs:
            raise KeyError('Content-type %r not registered for %r' % (
                content_type, self
            ))

        return make_response(
            self.content_funcs[content_type](data, content_type, status)
        )

    def copy(self):
        return deepcopy(self)


content_renderer = PluggableRenderer()
exception_renderer = PluggableRenderer()



@content_renderer.renders('application/json')
def render_json_content(data, content_type, status):
    return jsonext.dumps(data), status, {'Content-type': content_type}


@content_renderer.renders('text/plain')
def render_text_plain_content(data, content_type, status):
    return pformat(data), status, {'Content-type': 'text/plain; charset=ascii'}


@exception_renderer.renders('text/plain')
def render_text_plain_exception(exc, content_type, status):
    # renders an exception as ascii text
    text = exc.description.encode(TEXT_PLAIN_ENCODING)

    return text, exc.code, {'Content-type': 'text/plain; charset=utf8'}


@exception_renderer.renders('text/html')
def text_html(exc, content_type, status):
    tpl_name = current_app.config.get('EXCEPTION_TEMPLATE_TEXT_HTML',
                                      'exception.html')
    tpl = current_blueprint.absolute_jinja_env.get_or_select_template(tpl_name)
    html = tpl.render(exc=exc)

    return html, exc.code, {'Content-type': 'text/html; charset=utf8'}


@exception_renderer.renders('application/problem+json')
@exception_renderer.renders('application/json')
def application_problem_json(exc, content_type, status):
    data = {
        'type': ('https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#%d'
                 % exc.code),
        'title': exc.name,
        'status': exc.code,
        'detail': exc.description,
    }

    return json.dumps(data), exc.code, {'Content-type':
                                        'application/problem+json'}



#!/usr/bin/env python

from flask import Blueprint, request, abort, make_response
from flask.helpers import locked_cached_property
from jinja2 import PackageLoader, ChoiceLoader, Environment
import werkzeug

from .helpers import get_best_mimetype, MIMEMap, register_converter
from .resources import ResourceView
from . import renderers

__version__ = '0.5.0'


class ContentNegotiationMixin(object):
    """A blueprint mixin that supports content negotiation. Used in conjunction
    with the :func:`.get_best_mimetype()` and :func:`.serialize_response()`
    functions.

    :attr:`~flask_arrest.ContentNegotiationMixin.incoming` types are checked
    whenever a client sends a request with content to this blueprint. If the
    supplied ``Content-type``-Header is not among the MIME-Types valid for
    the specific endpoint, the request is rejected with a
    :py:class:`~werkzeug.exceptions.UnsupportedMediaType` exception.

    :attr:`~flask_arrest.ContentNegotiationMixin.outgoing` types supply
    information about which possible mimetypes can be sent back to the client.
    Renderers for data can use these to find an intersection with the
    ``Accept``-headers the client sent. Many will send an HTTP 406 (Not
    Acceptable) error if none of the advertised types is found in the clients
    ``Accept``-header."""

    def __init__(self, *args, **kwargs):
        super(ContentNegotiationMixin, self).__init__(*args, **kwargs)
        self.before_request(self.__check_incoming_content_type)

        self.incoming = MIMEMap()
        """a :py:class:`~flask_arrest.helpers.MIMEMap` of incoming data types.
        The default will contain just ``application/json``."""

        self.incoming.add_mimetype('application/json')

        self.outgoing = MIMEMap()
        """a :py:class:`~flask_arrest.helpers.MIMEMap` of outgoing data types.
        The default will contain just ``application/json``."""
        self.outgoing.add_mimetype('application/json')

    def __check_incoming_content_type(self):
        if not request.content_type and (request.data or request.form):
            # client needs to send a content-type, if he sends
            # content
            abort(415)

        if not request.content_type and not (request.data or request.form):
            return  # no content, no problem

        # strip the blueprint prefix
        prefix = self.name + '.'
        if request.endpoint.startswith(prefix):
            endpoint_name = request.endpoint[len(prefix):]
        else:
            endpoint_name = request.endpoint
        accepted = self.incoming.get_mimetypes(endpoint_name)
        if request.content_type not in accepted:
            abort(415)


class AbsoluteJinjaEnvMixin(object):
    """Jinja environment helper mixin.

    Creates a separate jinja_env for temlates that come bundled with a
    blueprint (when deriving from a blueprint, normally the template_path will
    be overridden).
    """

    @locked_cached_property
    def _absolute_jinja_loader(self):
        # we override this so we can add our own template path as well as the
        # the one of any deriving blueprint
        #
        # templates are stored in flask_arrest/templates
        exc_loader = PackageLoader(__name__.rsplit('.', 1)[0])
        if not self.jinja_loader:
            return exc_loader

        return ChoiceLoader([self.jinja_loader, exc_loader])

    @locked_cached_property
    def absolute_jinja_env(self):
        env = Environment(loader=self._absolute_jinja_loader)

        return env


# FIXME: this may or may not be removed
class ResourceMountMixin(object):
    def mount_resource(self, handler):
        # NOTE: we are not using converters to unmarshal right now - exceptions
        #       triggered by loading resources through converters will not
        #       get handled by the blueprint exception handlers. this may
        #       or may not be possible to remedy by registering application-
        #       wide exceptions handlers
        #
        # create a converter for handler
        class Converter(werkzeug.routing.BaseConverter):
            def to_python(self, value):
                return value

            def to_url(self, obj):
                return handler._obj_to_id(obj)

        register_converter(self, handler.singular, Converter)

        # note: we use getattr instead of hasattr to allow classes to override
        #       methods they don't want with None to hide them
        for target, data in handler.uris.items():
            name = ResourceView.construct_endpoint(handler, target, *data)

            if getattr(handler, target, None):
                self.add_url_rule(
                    data[1].format(handler),
                    view_func=ResourceView.as_view(name, handler),
                    methods=data[0])


class RestBlueprint(AbsoluteJinjaEnvMixin, ContentNegotiationMixin,
                    ResourceMountMixin, Blueprint):
    """A REST Blueprint."""

    def __init__(self, *args, **kwargs):
        super(RestBlueprint, self).__init__(*args, **kwargs)

        self.http_errorhandlers(self.__serializing_errorhandler)

        self.content_renderer = renderers.content_renderer.copy()
        """The content renderer to use as the default. Usually called by
        :py:func:`~flask_arrest.helpers.serialize_response`, should support
        the :py:class:`~flask_arrest.renderers.Renderer` interface.

        Per default, a copy of
        :py:attr:`~flask_arrest.renderers.content_renderer` is used as the
        initial value."""

        self.exception_renderer = renderers.exception_renderer.copy()
        """The exception renderer that is used to render every
        :py:class:`~werkzeug.exceptions.HTTPException` thrown inside this
        blueprint. Should  support the
        :py:class:`~flask_arrest.renderers.Renderer` interface."""

    def http_errorhandlers(self, f):
        """Decorator for registering a function as an exception handler
        for all instances of :py:class:`~werkzeug.exceptions.HTTPException`.

        This function will go away as soon as the following
        issue in flask is fixed: https://github.com/mitsuhiko/flask/pull/952"""
        # there's an issue and a pull request for this at
        # https://github.com/mitsuhiko/flask/pull/952
        # for now, this is a workaround
        # ideally, we would just myblueprint.errorhandler(HTTPException)(f)

        # as of Flask 0.11, the old code threw an exception upon construction,
        # however it seems to be still not possible to register an exception
        # for the base class, because http error codes seem to be checked
        # numerically instead of by type
        for i in werkzeug.exceptions.default_exceptions.keys():
            self.register_error_handler(i, f)
        return f

    def __serializing_errorhandler(self, exc):
        content_type = get_best_mimetype()

        if not content_type:
            # we found no acceptable content-type to render the exception
            # return nothing but the code
            code = getattr(exc, 'code', 500)
            return make_response('', code, {})

        return self.exception_renderer.render_response(exc, content_type)



from flask import request
from flask.views import View
from werkzeug.exceptions import NotFound

from .helpers import serialize_response


class ResourceView(View):
    VIEW_DELIM = ':'

    def __init__(self, handler):
        self.handler = handler

    def dispatch_request(self, *args, **kwargs):
        action = getattr(self.handler,
                         self.extract_endpoint_target(request.endpoint))
        return serialize_response(action(*args, **kwargs))

    @classmethod
    def construct_endpoint(cls, handler, target,
                           methods, uri_template, override=None):
        if override:
            return override.format(handler)
        return handler.singular + cls.VIEW_DELIM + target

    @classmethod
    def extract_endpoint_target(cls, name):
        if cls.VIEW_DELIM in name:
            return name.rsplit(cls.VIEW_DELIM, 1)[1]
        else:
            # no VIEW_DELIM in endpoint name, default to show
            return 'show'


class HandlerMixin(object):
    uris = {
        'show': (['GET'], '/{0.singular}/<{0.singular}:obj_id>/',
                 '{0.singular}'),
        'create': (['POST'], '/{0.plural}/'),
        'update': (['PATCH'], '/{0.singular}/<{0.singular}:obj_id>/'),
        'replace': (['PUT'], '/{0.singular}/<{0.singular}:obj_id>/'),
        'remove': (['DELETE'], '/{0.singular}/<{0.singular}:obj_id>/'),
        'query': (['GET'], '/{0.plural}/'),
    }

    def _obj_to_id(self, obj):
        return str(obj.id)

    def show(self, obj_id):
        try:
            obj = self._from_id(obj_id)
        except (ValueError, KeyError):
            raise NotFound()
        return obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='Flask-arrest',
      version='0.5.0',
      description='A small Flask extension to ease the creation of REST apis.',
      long_description=read('README.rst'),
      author='Marc Brinkmann',
      author_email='git@marcbrinkmann.de',
      url='http://github.com/mbr/Flask-arrest',
      license='MIT',
      packages=find_packages(exclude=['test']),
      include_package_data=True,
      zip_safe=False,
      install_requires=['Flask', 'werkzeug', 'jsonext'],
      tests_require=['pytest'],
      classifiers=[
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
      ])
